// -- this file should contain all the methods used to create the initial simulation data

package shared.model;

import shared.model.structure.*;
import shared.model.storage.*;

import java.util.*;
import javax.vecmath.*;

class SimulationCreator
{
        private int numObjects = 0;
        private SimulationStore storage;

        SimulationCreator(SimulationStore s)
        {
                storage = s;
        }

        ////////// Register templates as part of the simulation //////////

        void addTemplate(Object template)
        {
                if (template instanceof ParticleTemplate)
                {
                        // store the particle template
                        storage.putTemplate(((ParticleTemplate) template).getId(), template);
                }

                if (template instanceof SimObjectTemplate)
                {

                        // store the object template
                        storage.putTemplate(((SimObjectTemplate) template).getId(), template);
                }

                if (template instanceof InteractionTemplate)
                {
                        // store the interaction template
                        storage.putTemplate(((InteractionTemplate) template).getId(), template);
                }

                if (template instanceof StateTemplate)
                {
                        // store the interaction template
                        storage.putTemplate(((StateTemplate) template).getId(), template);
                }

                if (template instanceof ConditionTemplate)
                {
                        // store the interaction template
                        storage.putTemplate(((ConditionTemplate) template).getId(), template);
                }

                if (template instanceof PotentialTemplate)
                {
                        storage.putTemplate(((PotentialTemplate) template).getId(), template);
                }
        }

        //////// Create instances of objects and return their hash strings //////////

        String createObjectInstance(SimObject obj)
        {
                return createObjectInstance(obj.getTemplateId(), "O" + (numObjects++), obj.getCentreOfMass());
        }

        private String createObjectInstance(String templateId, String objectHash, Vector3f com)
        {
                System.out.println("create object instance");
                // create the object instance
                SimObjectTemplate objTemplate = (SimObjectTemplate) storage.getTemplate(templateId);
                SimObject object = new SimObject(objectHash);
                object.setTemplateId(templateId);
                object.setCentreOfMass(com);

                // create the particle instances
                String[] particleTemplateIds = objTemplate.getParticleTemplateIds();
                String[] particleIds = new String[particleTemplateIds.length];
                Vector3f[] particleCOM = objTemplate.getParticleCentreOfMasses();
                for (int i=0; i<particleTemplateIds.length; i++)
                {

                        Vector3f newCOM = new Vector3f();
                        newCOM.x = com.x + particleCOM[i].x;
                        newCOM.y = com.y + particleCOM[i].y;
                        newCOM.z = com.z + particleCOM[i].z;
                        particleIds[i] = createParticleInstance(particleTemplateIds[i], objectHash + ".P" + i, newCOM);
                }
                object.setParticleIds(particleIds);

                // create the subobject instances
                String[] subObjectTemplateIds = objTemplate.getSubObjectTemplateIds();
                String[] subObjectIds = new String[subObjectTemplateIds.length];
                for (int i=0; i<subObjectTemplateIds.length; i++)
                {
                        subObjectIds[i] = createObjectInstance(subObjectTemplateIds[i], objectHash + ".O" + i, com);
                }
                object.setSubObjectIds(subObjectIds);

                storage.putInstance(objectHash, object);

                // create the interaction instances
                String[] stateTemplateIds = objTemplate.getStateTemplateIds();
                String[] stateIds = new String[stateTemplateIds.length];
                for (int i=0; i<stateIds.length; i++)
                {
                        stateIds[i] = createStateInstance(stateTemplateIds[i], objectHash, ".S" + i);
                }
                object.setStateIds(stateIds);

                return objectHash;
        }

        private String createStateInstance(String templateId, String objectHash, String extension)
        {
                System.out.println("create state instance");

                String stateHash = objectHash + extension;

                State state = new State(stateHash);
                state.setTemplateId(templateId);
                StateTemplate template = (StateTemplate) (storage.getTemplate(templateId));

                // create the interaction instances
                String[] interactionTemplateIds = template.getInteractionTemplateIds();
                String[] interactionIds = new String[interactionTemplateIds.length];
                for (int i=0; i<interactionIds.length; i++)
                {
                        interactionIds[i] = createInteractionInstance(interactionTemplateIds[i], objectHash, extension + ".I" + i);
                }
                state.setInteractionIds(interactionIds);

                // create the condition instances
                String[] conditionTemplateIds = template.getConditionTemplateIds();
                String[] conditionIds = new String[conditionTemplateIds.length];
                for (int i=0; i<conditionIds.length; i++)
                {
                        conditionIds[i] = createConditionInstance(conditionTemplateIds[i], objectHash, extension + ".C" + i);
                }
                state.setConditionIds(conditionIds);

                storage.putInstance(stateHash, state);
                return stateHash;
        }

        private String createConditionInstance(String templateId, String objectHash, String conditionExt)
        {

                String conditionHash = objectHash + conditionExt;
                Condition condition = new Condition(conditionHash);
                condition.setTemplateId(templateId);

                ConditionTemplate template = (ConditionTemplate) (storage.getTemplate(templateId));
                String[] conditionRelativePaths = template.getConditionComponents();
                String[] conditionAbsolutePaths = new String[conditionRelativePaths.length];
                for (int i=0; i<conditionRelativePaths.length; i++)
                {
                        conditionAbsolutePaths[i] = objectHash + "." + conditionRelativePaths[i];
                }
                condition.setConditionComponents(conditionAbsolutePaths);

                storage.putInstance(conditionHash, condition);

                return conditionHash;
        }

        private String createInteractionInstance(String templateId, String objectHash, String interactionExt)
        {
                System.out.println("create interaction instance");

                String interactionHash = objectHash + interactionExt;
                Interaction interaction = new Interaction(interactionHash);
                interaction.setTemplateId(templateId);

                InteractionTemplate template = (InteractionTemplate) (storage.getTemplate(templateId));

                // set the component paths
                String[] interactRelativePaths = template.getComponentPaths();
                String[] interactAbsolutePaths = new String[interactRelativePaths.length];
                for (int i=0; i<interactRelativePaths.length; i++)
                {
                        interactAbsolutePaths[i] = objectHash + "." + interactRelativePaths[i];
                }
                interaction.setComponentPaths(interactAbsolutePaths);

                // set the potentials
                String[] potentialTemplates = template.getPotentialTemplates();
                String[] forceMults = template.getForceMults();
                String[] valueMults = template.getValueMults();

                String[] potentials = new String[potentialTemplates.length];

                for (int i=0; i<potentialTemplates.length; i++)
                {
                        potentials[i] = createPotentialInstance(potentialTemplates[i], objectHash + interactionExt+".T"+i,
                                                                forceMults[i], valueMults[i]);
                }
                interaction.setPotentials(potentials);

                storage.putInstance(interactionHash, interaction);

                return interactionHash;
        }

        private String createPotentialInstance(String templateId, String potentialHash, String forceMult, String valueMult)
        {
                float forceMultiplier = 0f;
                float valueMultiplier = 0f;
                try
                {
                        forceMultiplier = Float.parseFloat(forceMult);
                        valueMultiplier = Float.parseFloat(valueMult);
                }
                catch(NumberFormatException e)
                {
                }

                PotentialTemplate potentialTemp = (PotentialTemplate) (storage.getTemplate(templateId));
                String[] forces = potentialTemp.getForces();
                String[] values = potentialTemp.getValues();

                for (int i=0; i<forces.length; i++)
                {
                        float force = 0f;
                        try
                        {
                                force = Float.parseFloat(forces[i]);
                        }
                        catch(NumberFormatException e)
                        {
                        }

                        force *= forceMultiplier;
                        forces[i] = ""+force;
                }

                for (int i=0; i<values.length; i++)
                {
                        float value = 0f;
                        try
                        {
                                value = Float.parseFloat(values[i]);
                        }
                        catch(NumberFormatException e)
                        {
                        }

                        value *= valueMultiplier;
                        values[i] = ""+value;
                }

                Potential potential = new Potential(potentialHash);
                potential.setForces(forces);
                potential.setValues(values);

                storage.putInstance(potentialHash, potential);

                return potentialHash;
        }

        private String createParticleInstance(String templateId, String particleHash, Vector3f com)
        {
                // create the particle instance
                Particle particle = new Particle(particleHash);
                particle.setTemplateId(templateId);
                particle.setCentreOfMass(com);

                storage.putInstance(particleHash, particle);

                return particleHash;
        }

        int getNumObjects()
        {
                return numObjects;
        }
}