// -- This file should provide an interface to the simulation data and contain all the methods which operate on that data
// -- (once the simulation has been created). The user should not call these methods.

package shared.model;

import shared.model.storage.*;
import shared.model.structure.*;

import java.util.*;
import javax.vecmath.*;

class SimulationManipulator
{
        private SimulationStore storage;
        private float simulationSize, halfSimulationSize;
        private Boundary boundary;

        SimulationManipulator(SimulationStore s)
        {
                storage = s;
                boundary = new Boundary();
        }

        // -- boundary operations
        public Vector3f getPosition(Vector3f pos)
        {
                return boundary.getPosition(pos);
        }

        public boolean testPosition(Vector3f pos)
        {
                return boundary.testPosition(pos);
        }

        public void setSimulationSize(float s)
        {
                boundary.setSimulationSize(s);
        }

        public void setBoundaryType(String type)
        {
                boundary.setType(type);
        }

        public float getSquaredDistance(Vector3f vect1, Vector3f vect2)
        {
                return boundary.getSquaredDistance(vect1, vect2);
        }

        public float getDistance(Vector3f vect1, Vector3f vect2)
        {
                return boundary.getDistance(vect1, vect2);
        }

        // -- single particle operations

        public float getParticleRadius(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                ParticleTemplate pt = (ParticleTemplate)(storage.getTemplate(p.getTemplateId()));
                return pt.getShapeRadius();
        }

        public float getParticleMass(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                ParticleTemplate pt = (ParticleTemplate)(storage.getTemplate(p.getTemplateId()));
                return pt.getMass();
        }

        public Vector3f translateParticle(String id, Vector3f trans)
        {
                Particle p = (Particle)(storage.getInstance(id));

                Vector3f com = p.getCentreOfMass();
                com.x += trans.x;
                com.y += trans.y;
                com.z += trans.z;
                Vector3f newPos = boundary.getPosition(com);
                p.setCentreOfMass(newPos);

                return newPos;
        }

        public Vector3f getParticleCentreOfMass(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                return p.getCentreOfMass();
        }

        public void setParticleCentreOfMass(String id, Vector3f com)
        {
                Particle p = (Particle)(storage.getInstance(id));
                p.setCentreOfMass(com);
        }

        public AxisAngle4f getParticleOrientation(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                return p.getOrientation();
        }

        public void setParticleOrientation(String id, AxisAngle4f orient)
        {
                Particle p = (Particle)(storage.getInstance(id));
                p.setOrientation(orient);
        }

        // -- multiple particle operations

        String[] getAllParticlesOfObject(String objId)
        {
                Vector vect = new Vector();
                getParticleIdsRecursiveVect(vect, objId);
                String[] ids = new String[vect.size()];
                for (int i=0; i<vect.size(); i++)
                {
                        ids[i] = (String) (vect.elementAt(i));
                }
                return ids;
        }

        public Vector getParticleIdsRecursiveVect(Vector vect, String objId)
        {
                SimObject obj = (SimObject) (storage.getInstance(objId));

                String[] particleIds = obj.getParticleIds();
                for (int i=0; i<particleIds.length; i++)
                {
                        vect.addElement(particleIds[i]);
                }

                String[] subObjectIds = obj.getSubObjectIds();
                for (int i=0; i<subObjectIds.length; i++)
                {
                        getParticleIdsRecursiveVect(vect, subObjectIds[i]);
                }
                return vect;
        }

        String[] getAllParticleIds()
        {
                Set set = storage.getInstanceEntrySet();
                Vector particles = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();

                        Object obj = me.getValue();

                        if (obj instanceof Particle)
                        {
                                particles.addElement(me.getKey());
                        }
                }

                String[] partIds = new String[particles.size()];
                for (int i=0; i<particles.size(); i++)
                {
                        partIds[i] = (String) (particles.elementAt(i));
                }

                return partIds;
        }

        // -- multiple object operations

        String[] getAllObjectIds()
        {
                Set set = storage.getInstanceEntrySet();
                Vector objects = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();

                        Object obj = me.getValue();

                        if (obj instanceof SimObject)
                        {
                                objects.addElement(me.getKey());
                        }
                }

                String[] objIds = new String[objects.size()];
                for (int i=0; i<objects.size(); i++)
                {
                        objIds[i] = (String) (objects.elementAt(i));
                }

                return objIds;
        }

        /*
        ParticleTemplate[] getAllParticleTemplates()
        {
                Set set = storage.getTemplateEntrySet();
                Vector particleTemplates = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();

                        Object obj = me.getValue();

                        if (obj instanceof ParticleTemplate)
                        {
                                particleTemplates.addElement(obj);
                        }
                }

                ParticleTemplate[] partTemps = new ParticleTemplate[particleTemplates.size()];
                for (int i=0; i<particleTemplates.size(); i++)
                {
                        partTemps[i] = (ParticleTemplate) (particleTemplates.elementAt(i));
                }

                return partTemps;
        }

        */

        Vector getAllTemplates()
        {
                Set set = storage.getTemplateEntrySet();
                Vector objectTemplates = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();
                        Object obj = me.getValue();
                        objectTemplates.addElement(obj);
                }

                return objectTemplates;
        }
}

class Boundary
{
        private String type = "walled";
        private float simulationSize, halfSimulationSize;

        Boundary()
        {
                setSimulationSize(1f);
        }

        void setSimulationSize(float simSize)
        {
                simulationSize = simSize;
                // -- compute half of sim size for efficiency
                halfSimulationSize = 0.5f * simSize;
        }

        void setType(String t)
        {
                type = t;
        }

        boolean testPosition(Vector3f c)
        {
                // switch for different boundaries
                if (type.equals("tesselated"))
                {
                        return true;
                }
                else
                {
                        // assume boundary = "walled"
                        return testPositionWalled(c);
                }
        }

        boolean testPositionWalled(Vector3f c)
        {
                if (c.x >= halfSimulationSize || c.x <= -halfSimulationSize) { return false; }
                if (c.y >= halfSimulationSize || c.y <= -halfSimulationSize) { return false; }
                if (c.z >= halfSimulationSize || c.z <= -halfSimulationSize) { return false; }
                return true;
        }

        Vector3f getPosition(Vector3f c)
        {
                // switch for different boundaries
                if (type.equals("tesselated"))
                {
                        return getPositionTesselated(c);
                }
                else
                {
                        // assume boundary = "walled"
                        return c;
                }
        }

        Vector3f getPositionTesselated(Vector3f c)
        {
                Vector3f com = new Vector3f(c);

                while (com.x >= halfSimulationSize) {com.x -= simulationSize;}
                while (com.x <= -halfSimulationSize){com.x += simulationSize;}
                while (com.y >= halfSimulationSize) {com.y -= simulationSize;}
                while (com.y <= -halfSimulationSize){com.y += simulationSize;}
                while (com.z >= halfSimulationSize) {com.z -= simulationSize;}
                while (com.z <= -halfSimulationSize){com.z += simulationSize;}

                return com;
        }

        float getDistance(Vector3f v1, Vector3f v2)
        {
                // switch for different boundaries if present - not at the moment
                return getDistanceTesselated(v1, v2);
        }

        float getDistanceTesselated(Vector3f v1, Vector3f v2)
        {
                return (float) Math.sqrt( getSquaredDistanceTesselated(v1, v2) );
        }

        float getSquaredDistance(Vector3f v1, Vector3f v2)
        {
                // switch for different boundaries if present - not at the moment
                return getSquaredDistanceTesselated(v1, v2);
        }

        float getSquaredDistanceTesselated(Vector3f v1, Vector3f v2)
        {
                Vector3f v3 = new Vector3f(v2.x - v1.x, v2.y - v1.y, v2.z - v1.z);
                v3 = getPositionTesselated(v3);
                return (v3.x*v3.x + v3.y*v3.y + v3.z*v3.z);
        }

}





