package shared.update;

import shared.model.SimulationModel;
import shared.model.structure.*;
import shared.model.utils.*;

import java.util.*;
import javax.vecmath.*;

public class Evolver
{
        SimulationModel simulation;

        final float BOLTZMANN_CONSTANT = 1.3807f * (float) Math.pow(10.0, -23.0);

        float diffusionCoefficient;
        Random rand = new Random();
        float timestep = 0f;
        float waterViscosity = 0f;
        float acceptanceOffset = 0f;
        float temperature = 300f;

        int totalChanges = 0;
        int acceptedChanges = 0;

        public Evolver(SimulationModel sm)
        {
                simulation = sm;
                timestep = 1f;
                monteCarloInit();
        }

        public void setTimestep(float ts)
        {
                timestep = ts;
                monteCarloInit();
        }

        public float getTimestep()
        {
                return timestep;
        }

        public void update()
        {
                float totalEnergy = 0f;
                for (int i=0; i<200; i++)
                {
                        totalEnergy += monteCarloUpdate();
                }
                System.out.println("E total = "+totalEnergy+"\tAcceptance = "+((100f * (float)acceptedChanges) / (float)totalChanges)+"%");
        }

        public void monteCarloInit()
        {
                waterViscosity = 0.001f; // pascal seconds.
                acceptanceOffset = 0.5f;
                temperature = 300f;
        }

        public float monteCarloUpdate()
        {
                // -- get a random object
                String[] objectIds = simulation.getRootObjects();
                int objectNo = rand.nextInt(objectIds.length);

                String[] particleIds = simulation.getAllParticlesOfObject(objectIds[objectNo]);

                // -- calculate a centre of mass and effective radius for the object
                Vector3f objectCOM = new Vector3f();
                float objectVol = 0f;
                float totalMass = 0f;
                for (int i=0; i<particleIds.length; i++)
                {
                        Vector3f particleCOM = simulation.getParticleCentreOfMass(particleIds[i]);
                        float particleMass = simulation.getParticleMass(particleIds[i]);
                        totalMass += particleMass;
                        objectCOM.x += particleCOM.x * particleMass;
                        objectCOM.y += particleCOM.y * particleMass;
                        objectCOM.z += particleCOM.z * particleMass;

                        float particleRadius = simulation.getParticleRadius(particleIds[i]);
                        objectVol += 4f * (float) Math.PI * particleRadius * particleRadius * particleRadius / 3f;
                }
                objectCOM.x /= totalMass;
                objectCOM.y /= totalMass;
                objectCOM.z /= totalMass;

                float objectRadius = (float) Math.pow(3f * objectVol / ((float) Math.PI * 4f), (1f / 3f));

                // -- compute the total energy the particles of the object contributes to the system - that includes the energy of other particles
                // -- At the moment this just involves computing the thermal energy (= kT/2)
                float energy = 0.5f * BOLTZMANN_CONSTANT * temperature;

                // -- compute the root-mean-square displacement of the object (currently assuming uniform diffusion)
                float dragCoefficient = 6f * (float) Math.PI * waterViscosity * objectRadius;
                diffusionCoefficient = BOLTZMANN_CONSTANT * temperature / dragCoefficient; // Einstein relation
                float rmsDisplacement = (float) Math.sqrt(2 * diffusionCoefficient * timestep);

                // -- compute the total energy due to the object in this new position
                float newEnergy = 0.5f * BOLTZMANN_CONSTANT * temperature;

                totalChanges++;

                // -- Monte Carlo algorithm body

                // -- do not update unless the new energy is lower than the old energy, or the exponential is less than a random amount
                if ((newEnergy <= energy) || ((float) Math.exp(-(newEnergy - energy) / (BOLTZMANN_CONSTANT * temperature))) > rand.nextFloat() + acceptanceOffset)
                {

                        float xDisp = rmsDisplacement * (float) (rand.nextGaussian());
                        float yDisp = rmsDisplacement * (float) (rand.nextGaussian());
                        float zDisp = rmsDisplacement * (float) (rand.nextGaussian());
                        Vector3f displacement = new Vector3f(xDisp, yDisp, zDisp);
                        Vector3f newPos = new Vector3f(xDisp+objectCOM.x, yDisp+objectCOM.y, zDisp+objectCOM.z);

                        String[] allParticles = simulation.getAllParticles();


                        // -- translate and apply a random rotation
                        //AxisAngle4f objectRotation = new AxisAngle4f(rand.nextFloat() - 0.5f, rand.nextFloat() - 0.5f, rand.nextFloat() - 0.5f,
                        //                                        rand.nextFloat() * 360f);
                        AxisAngle4f objectRotation = new AxisAngle4f(1f,0f,0f,3.141f / 180f);

                        // -- store the new centres of mass for each particle so that collisions can be tested for before
                        // -- applying the changes.
                        Vector3f[] finalPos = new Vector3f[particleIds.length];

                        for (int i=0; i<particleIds.length; i++)
                        {
                                // translate the particle to the origin.
                                Vector3f pCOM = simulation.getParticleCentreOfMass(particleIds[i]);
                                Vector3f transToOrigin = new Vector3f(pCOM.x - objectCOM.x, pCOM.y - objectCOM.y, pCOM.z - objectCOM.z);
                                // rotate the particle about the object's centre of mass
                                Matrix4f rotMatrix = new Matrix4f();
                                rotMatrix.set(objectRotation);
                                Vector3f rotAtOrigin = new Vector3f();
                                rotMatrix.transform(transToOrigin, rotAtOrigin);
                                // translate back
                                finalPos[i] = new Vector3f(rotAtOrigin.x + newPos.x, rotAtOrigin.y + newPos.y, rotAtOrigin.z + newPos.z);
                        }

                        boolean success = false;
                        // -- check the new position for collisions
                        for (int i=0; i<particleIds.length; i++)
                        {
                                // -- check the boundary
                                success |= simulation.testPosition(finalPos[i]);
                        }

                        if (success)
                        {
                                // rotate the individual particles and set the new positions.
                                for (int i=0; i<particleIds.length; i++)
                                {
                                        //simulation.translateParticle(particleIds[i], displacement);
                                        AxisAngle4f partRotation = simulation.getParticleOrientation(particleIds[i]);
                                        simulation.setParticleOrientation(particleIds[i], Functions.combine(objectRotation, partRotation));
                                        simulation.setParticleCentreOfMass(particleIds[i], finalPos[i]);
                                }

                                acceptedChanges++;
                        }
                }

                // calculate the total energy
                float totalEnergy = 0f;
                for (int i=0; i<particleIds.length; i++)
                {
                        Vector3f position = simulation.getParticleCentreOfMass(particleIds[i]);
                        totalEnergy += 0.5f * BOLTZMANN_CONSTANT * temperature;
                }
                return totalEnergy;
        }
}

class InteractionParticle
{
        String particleId;
        Interaction interaction;

        InteractionParticle(String id, Interaction inter)
        {
                particleId = id;
                interaction = inter;
        }

        String getId()
        {
                return particleId;
        }

        Interaction getInteraction()
        {
                return interaction;
        }
}